/* arith.c                                        *
*                                                 *
*  (c) Michael Schindler, all right reserved      *
*  Feb. 1997                                      *
*  http://eiunix.tuwien.ac.at/~michael            *
*                                                 *
*  arithmetic coder without bit-stuff (FAST!)     *
*  There is NO DOCUMENTATION and NO WARRANTY      *
*  For a general idea on arithmetic coders check  *
*  A. Moffats paper at the DCC95.                 *
*                                                 *
*  The basic idea here is to go from base 2 to    *
*  base 256 numbers.                              *
*                                                 *
*  Austrian law forced me to file a patent for    *
*  arithmetic coders with max_range/min_range > 3 *
*                                                 *
*  define NOWARN to suppress warnings about       *
*  bytes_to_follow overflows (faster)            */

#define NOWARN

#include "uint.h"
#include "arith.h"
#include <stdio.h>

/* SIZE OF ARITHMETIC CODE VALUES. */

#define CODE_BITS 31                   /* Number of bits in a code value   */
typedef uint4 code_value;              /* Type of an arithmetic code value */
                                       /* must accomodate CODE_BITS+1 bits */
/* it is highly recommended that the total frequency count is less than    */
/* 1 << (CODE_BITS-13) to minimize rounding effects.                       */


#define SHIFT_BITS (CODE_BITS - 8)
#define EXTRA_BITS ((CODE_BITS-1) % 8 + 1)
#define Top_value ((code_value)1 << CODE_BITS)
#define Bottom_value (Top_value >> 8)


static unsigned char enbuffer, debuffer;  /* char buffer                   */
static code_value enlow, enrange;         /* Encoder state                 */
static uint4 bytes_to_follow;             /* number of bytes to follow     */
static code_value deoffs, der, derange;   /* Decoder state                 */


static inline void outbyte()
{  unsigned char carry = enlow >> CODE_BITS;
   putchar( enbuffer + carry );
   carry += 0xff;                      /* carry = carry? 0x00 : 0xff       */
   while (bytes_to_follow) {
      putchar( carry );
      bytes_to_follow --;
   }
}


/* START ENCODING A STREAM OF SYMBOLS. c preceeds the coded message and is */
/* returned by start_decoding                                              */
void start_encoding( char c )
{  enlow = 0;                                 /* Full code range.          */
   enrange = Top_value;
   enbuffer = c;
   bytes_to_follow = 0;                       /* No bytes to follow        */
}

/* START DECODING A STREAM OF SYMBOLS. returns the char passed to          */
/* start_encoding() or EOF                                                 */
int start_decoding( void )
{  int c = getchar();
   debuffer = getchar();
   deoffs = debuffer >> (8-EXTRA_BITS);
   derange = 1 << EXTRA_BITS;
   return c;
}

/* ENCODE A SYMBOL. */
inline void encode_freq( freq sy_f, freq lt_f, freq tot_f )
{  code_value r, tmp;
   r = enrange / tot_f;
   tmp = r * lt_f;
   enlow += tmp;
   if (lt_f+sy_f < tot_f)
      enrange = r * sy_f;
   else
      enrange -= tmp;
   while (enrange <= Bottom_value) {
      unsigned char byte = enlow >> SHIFT_BITS;
      if (byte == 0xff)
         bytes_to_follow ++;
      else {
         outbyte();
         enbuffer = byte;
      }
      enlow = (enlow << 8) & (Top_value-1);
      enrange <<= 8;
   }
#ifndef NOWARN
   if (bytes_to_follow > 0x7ffffffbL) {
      fprintf(stderr,"Too many bytes outstanding - File too large\n");
      exit(1);
   }
#endif
}


/* DECODE THE NEXT SYMBOL. This includes the updating! */
inline symb decode_symbol( cumtbl *tbl )
{   symb s;                    /* Symbol decoded                           */
    s = getsym(decode_culfreq(tbl->totfreq), tbl); /* Find the next symbol */
    decode_update(tbl->f[s], getcf(s,tbl), tbl->totfreq);
    return s;
}

/* CALCULATE CULMULATIVE FREQUENCY FOR NEXT SYMBOL. This does no updating! */
inline freq decode_culfreq( freq tot_f )
{  freq tmp;
   while (derange <= Bottom_value) {
      deoffs = (deoffs<<8) | ((debuffer<<EXTRA_BITS)&0xff);
      debuffer = getchar();
      deoffs |= debuffer >> (8-EXTRA_BITS);
      derange <<= 8;
   }
   der = derange/tot_f;
   tmp = deoffs/der;
   return (tmp>=tot_f ? tot_f-1 : tmp);
}

/* Update decoding variables using the symbol frequency, the frequency of  */
/* <= symbols and the total frequency. Both decodes may be used intermixed */
inline void decode_update(freq sy_f, freq lt_f, freq tot_f)
{  code_value tmp;
   tmp = der * lt_f;
   deoffs -= tmp;
   if (lt_f + sy_f < tot_f)
     derange = der * sy_f;
   else
     derange -= tmp;
}


/* FINISH ENCODING THE STREAM. */
void done_encoding( void )
{  enlow += enrange >> 1;                      /* select middle of interval*/
   outbyte();                                  /* output outstanding bytes */
   putchar((char)(enlow >> SHIFT_BITS));
}

/* FINISH DECODING THE STREAM. */
void done_decoding( void )
{
}
