/* bwtmodel.c                                                             **
** This probability model was specifically developed for block-sort       **
** transformed data.                                                      **
** It encodes the data in the following way:                              **
** first there is a symbol of range 0-2 that shows which of the following **
** 3 models was used. The probabilities are taken from the last CACHESIZE **
** uses.                                                                  **
** The second item is the symbol, encoded using one of the three models:  **
** model 0 is a FIFO cache of the last CACHESIZE (8) symbols seen. The    **
** probability of the most recent symbol is set to zero temporarily.      **
** model 1 is a move-to-front cache of the last MTFSIZE (16) symbols      **
** thrown out of cache (model 0).                                         **
** model 2 is a full background model, but the probabilities of symbols   **
** currently encoded by the other models is temporarily set to zero.      **
** The third item is a binary run-length code using three symbols: 0, 1   **
** and end-of-code. The leading 1 is omitted. This is equivalent to       **
** David Wheelers zero encoding after a MTF operation.                    **
*
** Reasons for these choices:                                             **
**   The small cache allows very fast adaptions to code changes; after a  **
** few symbol runs everything is up to date. The cache has the advantage  **
** over a MTF system that it maintains the individual symbol frequencies. **
** This is most efficient if there are just a few symbols used in the     **
** region; if there are just 2 symbols used there is no output at all.    **
**   The following MTF buffer is to assign short codes to a few recent    **
** symbols. This is most efficient for regions where more than just a few **
** symbols are used; for example after space and within numbers.          **
**   The background model provides a method to encode any other symbol, a **
** special endmark is also included to determine the end of file.         **
**   The method for run-length encoding was chosen to cope with the very  **
** varying run-lengths in BWT or ST transformend files. Wether the average**
** runlength needs 10 bits or 5 bits to represent makes no big difference **
** in symbol distribution with this method. I admit there are better      **
** encodings than plain binary, but there is no need to use that now :)   **
**   The choice to explicit encode what model was used instead of escapes **
** cooperates better with the cache buffer and minimizes loss for random  **
** files. With random files the cache and the MTF list will seldom encode **
** a symbol, whereas the full model will often encode symbols. This       **
** directly leads to a very short code for using the full model and to    **
** long codes for using the other models, so the average codelength will  **
** be just slightly above 8 bit.                                          **

** LEGAL MATTERS ======================================================== **

** I (Michael Schindler) wrote this software; parts of it (arith.c) are   **
** based on software by Mark Nelson. The method to calculate the indices  **
** in the b_i_t is from Peter Fenwick. The b_i_t is from Knuths           **
** fundamental alorithms.                                                 **

** I grant the right to use this software for research purposes provided  **
** that this legal matters section between LEGAL MATTERS and the (c) line **
** is kept intact (in code or as a seperate file); any other use prior to **
** Jan. 2000 requires witten permission. Email a brief description of the **
** intended use to michael@eiunix.tuwien.ac.at  Policy is to grant free   **
** permission for any non-commercial application.                         **

** Please note that the arithmetic coding used might be covered by patent **
** in your country. This is not the case in Austria, since at the time    **
** arithmetic coding was developed software was explicitely excluded from **
** patenting. When algorithms became patentable this was state of the art.**

** This software contains no known bugs, but: NO WARRANTY OF ANY KIND!!   **

** (c) Michael Schindler, Feb. 1997                                       */

#include <stdio.h>
#include "uint.h"
#include "model.h"
#include "b_i_t.h"
#include "arith.h"
#include <math.h>

#define ALPHABET 256
#define CACHESIZE 30
#define MTF1SIZE 10
#define MTF2SIZE 20

typedef struct ce {
   uint sym;       /* symbol */
   uint freq;      /* frequency of this symbol in cache */
   uint rlemodel;  /* rlemodel to use for this symbol */
   struct ce *next, *prev;
} cacheentry;

static cumtbl full, mtf1, mtf2;
/* full contains ALPHABET+1 symbols; frequency is set to 1 initially */
/* mtf1 contains MTF1SIZE symbols; frequency is set to 0 initially   */
/* mtf2 contains MTF2SIZE symbols; frequency is set to 0 initially   */

static cacheentry cache[CACHESIZE];

static uint mwhat[4],            /* model for what model to use      */
            mrun[4][3],          /* 4 run-length models              */
            encoding,            /* =1 if encoding;  =0 if decoding  */
            sym2mod[ALPHABET+1], /* model to encode the symbol       */
            mtf1size,            /* number of symbols in MTF1 list   */
            mtf2size;            /* number of symbols in MTF2 list   */
static cacheentry *sym2cache[ALPHABET+1], /* mapping symbol -> cache */
            *lastseen;           /* most recent symbol in cache      */
static symb mtf1list[MTF1SIZE],  /* sorted list of symbols in MTF1   */
            mtf2list[MTF2SIZE];  /* sorted list of symbols in MTF2   */
static uint4 runlength;          /* rulength of the last symbol      */


#define FULL_MAX  8000
#define FULL_INC     4
#define RUN_MAX    100
#define MTF_MAX    1000
#define MTF_INC      1
#define WHAT_INC     1
#define WHAT_SUM (mwhat[0]+mwhat[1]+mwhat[2]+mwhat[3])
#define WHAT_MAX    100

static void encoderun(uint4 n, uint *rmod)
{  uint sum = rmod[0] + rmod[1] + rmod[2];
   while (n>1) {
      if(n&1) {
         encode_freq(rmod[1]++, rmod[0], sum++);
      } else {
         encode_freq(rmod[0]++, 0, sum++);
      }
      n >>= 1;
   }
   encode_freq(rmod[2]++, rmod[0]+rmod[1], sum++);
   if (sum > RUN_MAX)
     for( sum=0; sum<3; sum++ )
        rmod[sum] = (rmod[sum]+1) >> 1;
}


static unsigned int decoderun(uint *rmod)
{  uint n=0, mask=1;
   uint sum = rmod[0] + rmod[1] + rmod[2];
   for (;;) {
      uint cf;
      cf = decode_culfreq( sum );
      if (cf<rmod[0])                                   /* symbol 0 */
         decode_update( rmod[0]++, 0, sum++);
      else {
         uint cf1;
         n |= mask;
         if (cf<(cf1=rmod[0]+rmod[1]))                  /* symbol 1 */
            decode_update( rmod[1]++, rmod[0], sum++);
         else {                                         /* symbol 2 */
            decode_update( rmod[2]++, cf1, sum++);
            if (sum > RUN_MAX)
              for( sum=0; sum<3; sum++ )
                 rmod[sum] = (rmod[sum]+1) >> 1;
            return n;
         }
      }
      mask <<= 1;
   }
}


static void cacheupdate( uint model, uint symbol, uint freq, uint rlemodel)
{ register cacheentry* ls;
   ls = lastseen->next;
   lastseen = ls;
   sym2cache[symbol] = ls;
   sym2mod[symbol] = 0;
   mwhat[model] += WHAT_INC;
   if (WHAT_SUM > WHAT_MAX) {
      uint i;
      for (i=0; i<4; i++)
         mwhat[i] = (mwhat[i]+1) >> 1;
   }
   /* remove last item in cache */
   if (ls->freq) {        /* last insance of this symbol in cache -> to MTF */
      uint i;
      if (mtf1size == MTF1SIZE) { /* MTF1 is complete -> last symbol to MTF2 */
         if (mtf2size == MTF2SIZE) { /* MTF2 is complete->last symbol to full*/
            symb s;
            s = mtf2list[MTF2SIZE-1];
            updatecumonly(s, &full, full.f[s]);
            sym2mod[s] = 3;
            if (full.totfreq > FULL_MAX) {
              scalefreqcond( &full, sym2mod );
            }
         } else {
            updatecumonly( mtf2size, &mtf2, mtf2.f[mtf2size] );
            mtf2size++;
         }
         for (i=mtf2size-1; i>0; i--)
            mtf2list[i] = mtf2list[i-1];
         mtf2list[0] = mtf1list[MTF1SIZE-1];
         sym2mod[mtf2list[0]] = 2;
      } else {
         updatecumonly( mtf1size, &mtf1, mtf1.f[mtf1size] );
         mtf1size++;
      }
      for (i=mtf1size-1; i>0; i--)
         mtf1list[i] = mtf1list[i-1];
      mtf1list[0] = ls->sym;
      sym2mod[ls->sym] = 1;
      ls->freq = freq;
   } else {
      ls->freq = freq;
      sym2cache[ls->sym]->freq--;
   }
   ls->sym = symbol;
   ls->rlemodel = rlemodel;
}

inline static void mtf1update( uint index )
{  uint i;
   updatefreq( index, &mtf1, MTF_INC );
   if ( mtf1.totfreq > MTF_MAX && mtf1size == MTF1SIZE )
      scalefreq( &mtf1 );
   mtf1size--;
   for (i=index; i<mtf1size; i++)
      mtf1list[i] = mtf1list[i+1];
   updatecumonly( mtf1size, &mtf1, -mtf1.f[mtf1size] );
}

inline static void mtf2update( uint index )
{  uint i;
   updatefreq( index, &mtf2, MTF_INC );
   if ( mtf2.totfreq > MTF_MAX && mtf2size == MTF2SIZE )
      scalefreq( &mtf2 );
   mtf2size--;
   for (i=index; i<mtf2size; i++)
      mtf2list[i] = mtf2list[i+1];
   updatecumonly( mtf2size, &mtf2, -mtf2.f[mtf2size] );
}

int initmodel(int mode)
{ int i;
  if (initfreq(&full, ALPHABET+1, 1))     /* initiaize full model        */
    return 1;
  mtf1size = 0;                           /* initiaize mtf1 model        */
  if (initfreq(&mtf1, MTF1SIZE, 0))
    return 1;
  for (i=0; i<MTF1SIZE; i++)
    mtf1.f[i] = 1;
  mtf2size = 0;                           /* initiaize mtf2 model        */
  if (initfreq(&mtf2, MTF2SIZE, 0))
    return 1;
  for (i=0; i<MTF2SIZE; i++)
    mtf2.f[i] = 1;
  for( i=0; i<4; i++)                     /* initialize runlength models */
     mrun[i][0] = mrun[i][1] = mrun[i][2] = 1;
  for( i=0; i<4; i++)                     /* initialize what model       */
     mwhat[i] = 1;
  lastseen = cache;                       /* initialize cache model      */
  for( i=0; i<CACHESIZE; i++) {
     cache[i].sym = ALPHABET;
     cache[i].freq = 0;
     cache[i].rlemodel = 3;
     cache[i].next = cache + i + 1;
     cache[i].prev = cache + i - 1;
  }
  cache[CACHESIZE-1].next = cache;
  cache[0].prev = cache+CACHESIZE-1;
  lastseen->freq = CACHESIZE;
  updatecumonly(ALPHABET, &full, -1);
  for( i=0; i<ALPHABET; i++)
     sym2mod[i] = 3;
  sym2mod[ALPHABET] = 0;
  sym2cache[ALPHABET] = lastseen;
  encoding = (mode!=0);
  if (encoding) {
    start_encoding((char)0);
    runlength = 1;
  } else {
    start_decoding();
    i=decoderun(mrun[lastseen->rlemodel]);
    runlength = 0;
  }
  return 0;
}

void deletemodel()
{ if (encoding) {
    encode(ALPHABET);
    done_encoding();
  } else
    done_decoding();
  freefreq(&mtf1);
  freefreq(&full);
}

void encode(int symbol)
{  if (symbol==lastseen->sym)
      runlength++;
   else {
      encoderun(runlength, mrun[lastseen->rlemodel]);
      lastseen->rlemodel = (runlength>2);
      if (sym2mod[symbol] == 0) {                  /* symbol in cache */
         cacheentry *tmp, *tmp1;
         uint sum = 0;
         encode_freq( mwhat[0], 0, WHAT_SUM);
         tmp = sym2cache[symbol];
         for (tmp1=tmp->next; tmp1 != lastseen; tmp1 = tmp1->next)
            sum += tmp1->freq;
         encode_freq( tmp->freq, sum, CACHESIZE-lastseen->freq );
         sum = tmp->freq + 1;
         tmp->freq = 0;
         cacheupdate(0, symbol, sum, tmp->rlemodel);
      } else if (sym2mod[symbol] == 1) {           /* symbol in MTF1  */
         uint i;
         encode_freq( mwhat[1], mwhat[0], WHAT_SUM);
         i=0;
         while (mtf1list[i]!=symbol)
            i++;
         encode_symbol( i, &mtf1);
         mtf1update(i);
         cacheupdate(1, symbol, 1, 2);
      } else if (sym2mod[symbol] == 2) {           /* symbol in MTF2  */
         uint i;
         encode_freq( mwhat[2], mwhat[0]+mwhat[1], WHAT_SUM);
         i=0;
         while (mtf2list[i]!=symbol)
            i++;
         encode_symbol( i, &mtf2);
         mtf2update(i);
         cacheupdate(2, symbol, 1, 2);
      } else {                                     /* symbol in full  */
         encode_freq( mwhat[3], mwhat[0]+mwhat[1]+mwhat[2], WHAT_SUM);
         encode_symbol( symbol, &full );
         updatecumonly( symbol, &full, -full.f[symbol] );
         full.f[symbol] += FULL_INC;
         cacheupdate(3, symbol, 1, 3);
      }
      runlength = 1;
   }
}

int decode(void)
{  if (runlength) {
      runlength--;
      return lastseen->sym;
   } else {
      symb s;
      int i;
      freq sum, tmp;
      tmp = decode_culfreq(WHAT_SUM);
      for (i=0,sum=0; (sum+=mwhat[i])<=tmp; i++)
         /* void */;
      decode_update(mwhat[i], sum-mwhat[i], WHAT_SUM);
      if (i==0) {                 /* cache */
         cacheentry *ce;
         tmp = decode_culfreq(CACHESIZE-lastseen->freq);
         for (ce = lastseen->prev, sum=0; (sum+=ce->freq)<=tmp; ce = ce->prev)
            /* void */;
         decode_update(ce->freq, sum-ce->freq, CACHESIZE-lastseen->freq);
         s = ce->sym;
         tmp = ce->freq + 1;
         ce->freq = 0;
         cacheupdate(0, s, tmp, ce->rlemodel);
      } else if (i==1){           /* mtf1  */
         tmp = decode_symbol(&mtf1);
         s = mtf1list[tmp];
         mtf1update(tmp);
         cacheupdate(1, s, 1, 2);
      } else if (i==2){           /* mtf2  */
         tmp = decode_symbol(&mtf2);
         s = mtf2list[tmp];
         mtf2update(tmp);
         cacheupdate(2, s, 1, 2);
      } else {                    /* full  */
         s = decode_symbol(&full);
         updatecumonly( s, &full, -full.f[s] );
         full.f[s] += FULL_INC;
         cacheupdate(3, s, 1, 3);
      } /* endif */
      if (s == ALPHABET)
         return EOF;
      runlength = decoderun(mrun[lastseen->rlemodel]) - 1;
      lastseen->rlemodel = (runlength>=2);
      return s;
   }
}
